/*
 * Instrumenter.java
 *
 * Created on June 21, 2001, 9:37 AM
 *
 * See LICENSE file for license conditions.
 */

package residue;

import de.fub.bytecode.generic.*;
import de.fub.bytecode.classfile.*;
import java.io.*;
import java.util.*;
import residue.events.*;
import residue.graph.*;
import residue.tables.*;
import residue.graph.Node;

import residue.runtime.Monitor;

/**
 * This class does the initial instrumentation of classes
 * which have no not been previously instrumented, or
 * have had their instrumentation removed with the Remover
 * class.
 *
 * @author  chowells
 * @version 
 */
public class Instrumenter extends AbstractInstrumenter
{
	private ConstantPoolGen cp;
	private int hit; // constant pool index for Monitor.hit()
	
	private String classname;
	private String sourcename;
	private String root;
	
	private Set cfgnames = new HashSet();
	
	private TableCreator table = new TableCreator();
	
	private String filename;
	private String methodname;
	
	private boolean reducenodes;
	private boolean breakloops;
	
	public static void main(String [] args) throws Exception
	{
		if (args.length < 2)
		{
			System.out.println("Please list files to instrument on the command line.");
			System.exit(1);
		}
		
		Instrumenter in = new Instrumenter();
		
		in.addInstrumentationEventListener(new InstrumentationEventListener()
		{
			public void instrumentationStarted(InstrumentationEvent e)
			{
				System.out.println("Beginning instrumentation");
			} // instrumentationStarted
			
			public void instrumentationFinished(InstrumentationEvent e)
			{
				System.out.println("Completed instrumentation");
			} // instrumentationStarted
			
			public void classStarted(InstrumentationEvent e)
			{
				System.out.println("Instrumenting class " + e.getClassName());
			} // instrumentationStarted
			
		}); // anonymous InstrumentationEventListener
		
		int i;
		if (args[0].equalsIgnoreCase("-breakloops"))
		{
			in.setReduceNodes(true);
			in.setBreakLoops(true);
			i = 2;
		}
		else if (args[0].equalsIgnoreCase("-noreduce"))
		{
			in.setReduceNodes(false);
			i = 2;
		}
		else
		{
			in.setReduceNodes(true);
			in.setBreakLoops(false);
			i = 1;
		}
		
		// Determine the path for storing tables in
		File p = new File(args[i - 1]);
		String path;
		if (p.isDirectory())
		{
			path = p.getCanonicalPath() + p.separatorChar;
		}
		else
		{
			System.out.println("The first path listed needs to be a directory");
			System.exit(1);
			path = null;
		}
		
		Set s = new HashSet();
		for (; i < args.length; i++)
			s.add(args[i]);
			
		in.instrument(s, path);
	} // main
	
	public void setReduceNodes(boolean value)
	{
		reducenodes = value;
	} // setReduceNodes
	
	public void setBreakLoops(boolean value)
	{
		breakloops = value;
	} // setBreakLoops
	
	public Set getConfigNames()
	{
		return Collections.unmodifiableSet(cfgnames);
	} // getConfigNames
	
	/**
	 * Instruments the classes given in the set s, which should be a set
	 * of String elements
	 */
	public void instrument(Set s, String path) throws IOException
	{
		instrumentationStarted(null, s.size());
		
		if (!path.endsWith(File.separator)) path += File.separator;
		
		Monitor.setLineTableName(path + "lineTable");
		Monitor.setHitTableName(path + "hitTable");
		
		int numbertogo = s.size() + 1;
		
		Iterator it = s.iterator();
		while (it.hasNext())
		{
			numbertogo--;
			filename = (String)it.next();
			
			if (!filename.endsWith(".class"))
			{
				System.out.println("unrecognized file type: " + filename);
				continue;
			}
			
			File f = new File(filename);
			
			if (!f.exists())
			{
				System.out.println("file not found: " + filename);
				continue;
			}
			
			classStarted(filename, numbertogo);
			
			filename = f.getCanonicalPath();
			
			JavaClass jc = new ClassParser(filename).parse();
			cp = new ConstantPoolGen(jc.getConstantPool());
			
			hit = cp.addMethodref("residue.runtime.Monitor", "hit", "(I)V");
			classname = jc.getClassName();
			
			String pack = jc.getPackageName().replace('.', File.separatorChar); 
			
			sourcename = pack + File.separatorChar + jc.getSourceFileName();
			root = filename.substring(0, filename.lastIndexOf(pack));
			
			// work in place
			Method [] m = jc.getMethods();
			for (int j = 0; j < m.length; j++)
				m[j] = instrument(m[j]);
			
			
			jc.setConstantPool(cp.getFinalConstantPool());
			jc.dump(filename);
		} // for
		
		// remove any existing hit table at the default name
		new File(Monitor.getHitTableName()).delete();
		
		Monitor.setSize(table.getHitIndex());
		
		it = cfgnames.iterator();
		while (it.hasNext())
		{
			String str = (String)it.next();
			
			Monitor.writeNames(str);
		}
		
		table.writeTo(Monitor.getLineTableName());
		
		instrumentationFinished(null, 0);
	} // instrument
	
	public Method instrument(Method m)
	{
		// sanity check.
		if (m.isNative() || m.isAbstract() || m.getCode() == null) return m;
		
		LineNumberTable lt = m.getCode().getLineNumberTable();
		
		// instrument the code
		MethodGen mg = new MethodGen(m, classname, cp);
		InstructionList il = mg.getInstructionList();
		
		InstructionHandle curr = il.getStart();
		
		// look for previously instrumented code and remove it
		while (curr != null)
		{
			if (curr.getInstruction() instanceof INVOKESTATIC)
			{
				INVOKESTATIC is = (INVOKESTATIC)curr.getInstruction();
				int init = cp.lookupMethodref("residue.runtime.Monitor", "init", "(Ljava/lang/String;)V");
				if (is.getIndex() == hit || is.getIndex() == init)
				{
					InstructionHandle oldtarget = curr.getPrev();
					InstructionHandle newtarget = curr.getNext();
					
					InstructionTargeter [] its = oldtarget.getTargeters();
					if (its != null)
						for (int i = 0; i < its.length; i++)
							its[i].updateTarget(oldtarget, newtarget);
					
					try
					{
						il.delete(oldtarget, curr);
					}
					catch (TargetLostException tle)
					{
						System.out.println("Target lost unexpected while removing previous instrumentation");
						System.out.println("File: " + filename);
						System.out.println("Method: " + methodname);
					}
					curr = newtarget;
				}
				else
				{
					curr = curr.getNext();
				}
			}
			else
			{
				curr = curr.getNext();
			}
		} // while
		
		methodname = mg.getName() + mg.getSignature();
		table.startMethod(filename, methodname, sourcename);
		table.addSourceRoot(root);
		
		CodeExceptionGen [] exes = mg.getExceptionHandlers();
		Graph g = new Graph(il, lt, exes, cp);
		
		// configure removal of nodes correctly
		Map removed;
		if (reducenodes)
			if (breakloops)
				removed = reduceBlocksConservative(g);
			else
				removed = reduceBlocksAggressive(g);
		else
			removed = new HashMap();
		
		// create BasicBlock -> Integer Map for endMethod
		Map refIndices = new HashMap();
		
		curr = il.getStart();
		while (curr != null)
		{
			BasicBlock bb = (BasicBlock)g.getBlockTable().get(curr);
			
			InstructionHandle first = bb.getFirstNode().getInstruction();
			InstructionHandle last = bb.getLastNode().getInstruction();
			
			if (!removed.containsKey(bb))
			{
				// create probe
				InstructionList probe = new InstructionList();
				probe.append(new PUSH(cp, table.getHitIndex()));
				probe.append(new INVOKESTATIC(hit));
				
				
				// insert before the last instruction in the block.
				il.insert(last, probe);
				probe.dispose();
				
				InstructionHandle probeHandle = last.getPrev().getPrev();
				
				table.addHitEntry(bb, first, probeHandle, last, bb.getLines());
				
				// retarget so that the probe is included
				InstructionTargeter [] its = last.getTargeters();
				if (its != null)
					for (int j = 0; j < its.length; j++)
						its[j].updateTarget(last, last.getPrev().getPrev());
			}
			else
			{
				int ref = table.addRefEntry(bb, first, last, bb.getLines());
				
				Integer r = new Integer(ref);
				
				// Set the back value for the proper children
				Iterator it = ((Set)removed.get(bb)).iterator();
				while (it.hasNext())
					refIndices.put(it.next(), r);
			}
			
			// advance to next BasicBlock
			curr = last.getNext();
		} // while
		
		// instrument main or <clinit> specially
		if ((methodname.equals("main([Ljava/lang/String;)V") || methodname.equals("<clinit>()V")))
		{
			int init = cp.addMethodref("residue.runtime.Monitor", "init", "(Ljava/lang/String;)V");
			
			InstructionHandle first = il.getStart();
			
			// add init instruction, giving package + class name
			InstructionList initsequence = new InstructionList();
			String initval = classname.replace('.', '/');
			
			initsequence.append(new PUSH(cp, initval));
			initsequence.append(new INVOKESTATIC(init));
			
			InstructionHandle inserted = il.insert(initsequence);
			InstructionTargeter [] its = first.getTargeters();
			
			// retarget instances of LineNumberGen only
			if (its != null)
				for (int i = 0; i < its.length; i++)
					if (its[i] instanceof LineNumberGen)
						its[i].updateTarget(first, inserted);
			
			cfgnames.add(filename.substring(0, filename.length() - 6));
		} // if
		
		// finish setting values of the LineTableEntries.
		il.setPositions();
		table.endMethod(refIndices);
		
		// finish the method
		mg.setMaxStack();
		Method result = mg.getMethod();
		il.dispose();
		return result;
	} // instrument
	
	/**
	 * The goal of this method is to reduce the number of
	 * blocks in a graph that need to be instrumented.  It returns
	 * a Map from BasicBlock to Set of BasicBlock.  The key set of
	 * the Map is the set of elements which do NOT need to be
	 * instrumented.  The associated set of blocks with each key
	 * is the the set of children which are instrumented (perhaps
	 * indirectly themselves) in place of the given block.
	 *
	 * This method places an additional restriction.  There
	 * should not be any loops which have all their nodes in
	 * the key set of the resulting Map.
	 */
	public static Map reduceBlocksConservative(Graph g)
	{
		// this map may be altered before returning it to fit the restriction
		Map result = g.predominatorTree();
		Map predominators = g.predominators();
		
		// find back edges, and break the tree at any node with one
		Iterator importantBlocks = new ArrayList(g.getBlocks()).iterator();
		while (importantBlocks.hasNext())
		{
			BasicBlock b = (BasicBlock)importantBlocks.next();
			Set doms = (Set)predominators.get(b);
			
			Iterator targets = b.getEdges().iterator();
			while (targets.hasNext())
			{
				BasicBlock t = (BasicBlock)targets.next();
				
				if (doms.contains(t)) result.remove(b);
			} // while
		} // while
		
		return result;
	} // reduceBlocksConservative
	
	
	/**
	 * The goal of this method is to reduce the number of
	 * blocks in a graph that need to be instrumented.  It returns
	 * a Map from BasicBlock to Set of BasicBlock.  The key set of
	 * the Map is the set of elements which do NOT need to be
	 * instrumented.  The associated set of blocks with each key
	 * is the the set of children which are instrumented (perhaps
	 * indirectly themselves) in place of the given block.
	 */
	public static Map reduceBlocksAggressive(Graph g)
	{
		return g.predominatorTree();
	} // reduceBlocksAggressive
	
} // class Instrumenter
